@@ -4,14 +4,14 @@
 # 13.06.2011, 10:54:19: initial commit
 #
 # TODO: add functionality for concurrent model
-# TODO: suppliying k as a variable name breaks down because it is propagated without evaluation
+# TODO: supplying k as a variable name breaks down because it is propagated without evaluation
 #       and so k is a list(<expression>, default.k)  where a vector is expected.
 ###############################################################################
 #' Penalized function-on-function regression
 #' 
 #' Implements additive regression for functional and scalar covariates and functional responses.
 #' This function is a wrapper for \code{mgcv}'s \code{\link[mgcv]{gam}} and its siblings to fit models of the general form \cr
-#' \eqn{E(Y_i(t)) = g(\mu(t) + \int X_i(s)\beta(s,t)ds + f(z_{1i}, t) + f(z_{2i}) + z_{3i} \beta_3(t) + \dots} )\cr
+#' \eqn{E(Y_i(t)) = g(\mu(t) + \int X_i(s)\beta(s,t)ds + f(z_{1i}, t) + f(z_{2i}) + z_{3i} \beta_3(t) + \dots }\cr
 #' with a functional (but not necessarily continuous) response \eqn{Y(t)}, response function \eqn{g},
 #' (optional) smooth intercept \eqn{\mu(t)}, (multiple) functional covariates \eqn{X(t)} and scalar covariates
 #' \eqn{z_1}, \eqn{z_2}, etc. 
@@ -35,9 +35,11 @@
 #' Functional random intercepts \eqn{B_{0g(i)}(t)} for a grouping variable \code{g} can be specified via \code{~s(g, bs="re")}),
 #' functional random slopes \eqn{u_i B_{1g(i)}(t)} in a numeric variable \code{u} via \code{~s(g, u, bs="re")}).\cr 
 #' The marginal spline basis used for the index of the 
-#' the functional response is specified via the global argument \code{bs.yindex}. If necessary, this can be overriden for any term by 
-#' supplying a \code{bs.yindex}-argument, e.g. \code{~s(x, bs.yindex=list(bs="tp", k=7)} would yield a tensor product spline for which
-#' the marginal basis for the index of the response are 7 cubic thin-plate spline functions.\cr 
+#' the functional response is specified via the \emph{global} argument \code{bs.yindex}. If necessary, this can be overriden for any specific term by 
+#' supplying a \code{bs.yindex}-argument to that term in the formula, e.g. \code{~s(x, bs.yindex=list(bs="tp", k=7))} would yield a tensor
+#' product spline for which
+#' the marginal basis for the index of the response are 7 cubic thin-plate spline functions overriding the global default for the basis and penalty on
+#' the index of the response given by the \emph{global} \code{bs.yindex}-argument .\cr 
 #' Use \code{~-1 + c(1) + ...} to specify a model with only a constant and no functional intercept. \cr
 #' 
 #' The functional response and functional covariates have to be supplied as n by <no. of evaluations> matrices, i.e. each row is one functional observation.
@@ -45,7 +47,7 @@
 #' into one long vector and the covariates are expanded/repeated correspondingly. This means the models get quite big fairly fast,
 #' since the effective number of rows in the design matrix is number of observations times number of evaluations of \eqn{Y(t)} per observation.\cr 
 #' 
-#' Note that pffr overrides the default identifiability constraints (\eqn{\sum_{i,t} \hat f(z_i, x_i, t) = 0) implemented in \code{mgcv} 
+#' Note that pffr overrides the default identifiability constraints (\eqn{\sum_{i,t} \hat f(z_i, x_i, t) = 0}) implemented in \code{mgcv} 
 #' for tensor product terms whose marginal terms include the index of the functional response \eqn{t}.  Instead, \eqn{\sum_i \hat f(z_i, x_i, t) = 0} for all \eqn{t}
 #' is enforced, so that effects varying over \eqn{t} can be interpreted as deviations from the global functional intercept.
 #' I recommend using centered scalar covariates for terms like
@@ -55,8 +57,14 @@
 #' @param formula a formula with special terms as for \code{\link[mgcv]{gam}}, with additional special terms \code{\link{ff}()} and \code{c()}
 #' @param yind a vector with length equal to the number of columns of the matrix of functional responses giving the vector of evaluation points \eqn{(t_1, \dots ,t_{G})}.
 #' 	If \code{formula} contains an \code{\link{ff}}-term which specifies \code{yind} this is used. If neither is given, \code{yind} is \code{1:ncol(<response>)}.   
-#' @param fitter the name of the function used to estimate the model. Defaults to \code{\link[mgcv]{gam}} if the matrix of functional responses has less than \code{2e5} data points
+#' @param algorithm the name of the function used to estimate the model. Defaults to \code{\link[mgcv]{gam}} if
+#' the matrix of functional responses has less than \code{2e5} data points
 #' 	 and to \code{\link[mgcv]{bam}} if not. "gamm" (see \code{\link[mgcv]{gamm}}) and "gamm4" (see \code{\link[gamm4]{gamm4}}) are valid options as well.  
+#' @param data an (optional) \code{data.frame} or a named list containing the data. 
+#' The functional response and functional covariates have to be supplied as n by <no. of evaluations> matrices, i.e. each row is one functional observation.
+#' The model is then fitted with the data in long format, i.e., the rows of the matrix of the functional response evaluations \eqn{Y_i(t)} are stacked
+#' into one long vector and the covariates are expanded/repeated correspondingly. This means the models get quite big fairly fast,
+#' since the effective number of rows in the design matrix is number of observations times number of evaluations of \eqn{Y(t)} per observation.\cr 
 #' @param method Defaults to \code{"REML"}-estimation, including of unknown scale. See \code{\link[mgcv]{gam}} for details.  
 #' @param bs.yindex a named (!) list giving the parameters for spline bases on the index of the functional response. 
 #'  Defaults to \code{list(bs="ps", k=5, m=c(2, 1))}, i.e. 5 cubic B-splines bases with first order difference penalty. 
@@ -64,13 +72,48 @@
 #'  Defaults to \code{list(bs="ps", k=20, m=c(2, 1))}, i.e. 20 cubic B-splines bases with first order difference penalty.  
 #' @param tensortype which typ of tenor product splines to use. One of "\code{\link[mgcv]{te}}" or "\code{\link[mgcv]{t2}}", defaults to \code{te}
 #' @param ... additional arguments that are valid for \code{\link[mgcv]{gam}} or \code{\link[mgcv]{bam}}. \code{weights, subset, offset} are not yet implemented!
-#' @return a fitted \code{pffr}-object, which is a \code{\link[mgcv]{gam}}-object with some additional information in an \code{pffr}-entry. If \code{fitter} is
+#' @return a fitted \code{pffr}-object, which is a \code{\link[mgcv]{gam}}-object with some additional information in an \code{pffr}-entry. If \code{algorithm} is
 #'   \code{"gamm"} or \code{"gamm4"}, only the \code{$gam} part of the returned list is modified in this way.
 #' @author Fabian Scheipl, Sonja Greven
+#' @seealso \code{\link[mgcv]{smooth.terms}} for details of \code{mgcv} syntax and available spline bases
+#' @references Ivanescu, A., Staicu, A.-M., Scheipl, F. and Greven, S. (2012). 
+#'  Penalized function-on-function regression. (submitted) 
+#' \url{http://biostats.bepress.com/jhubiostat/paper240/}
+#' 
+#'  Scheipl, F., Staicu, A.-M. and Greven, S. (2012). Additive Mixed Models for Correlated Functional Data.
+#'  (submitted) \url{http://arxiv.org/abs/1207.5947}
+#' @examples  
+#' ###############################################################################
+#' # univariate model: 
+#' # Y(t) = f(t)  + \int X1(s)\beta(s,t)ds + eps  
+#' data1 <- pffrSim(scenario="1")
+#' t <- attr(data1, "yindex")
+#' s <- attr(data1, "xindex")
+#' m1 <- pffr(Y ~ ff(X1, yind=t, xind=s), data=data1)
+#' summary(m1)
+#' plot(m1, pers=TRUE)
+#' 
+#' ###############################################################################
+#' # multivariate model: 
+#' # Y(t) = f0(t)  + \int X1(s)\beta1(s,t)ds + \int X2(s)\beta2(s,t)ds +
+#' #		xlin \beta3(t) + f1(xte1, xte2) + f2(xsmoo, t) + beta4 xconst + eps  
+#' data2 <- pffrSim(scenario="2", n=200)
+#' t <- attr(data2, "yindex")
+#' s <- attr(data2, "xindex")
+#' m2 <- pffr(Y ~  ff(X1, xind=s, yind=t) + #linear function-on-function		
+#'                 ff(X2, xind=s, yind=t) + #linear function-on-function
+#'                 xlin  +  #varying coefficient term
+#'                 c(te(xte1, xte2)) + #bivariate smooth term in xte1 & xte2, const. over Y-index
+#'                 s(xsmoo) + #smooth effect of xsmoo varying over Y-index
+#'                 c(xconst), # linear effect of xconst constant over Y-index
+#'         data=data2)
+#' summary(m2)
+#' plot(m2, pers=TRUE)
 pffr <- function(
 		formula,
 		yind,
-		fitter = NA, 
+        data=NULL,
+		algorithm = NA, 
         method="REML",
         tensortype = c("te", "t2"),
 		bs.yindex = list(bs="ps", k=5, m=c(2, 1)), # only bs, k, m are propagated...
@@ -88,7 +131,7 @@ pffr <- function(
     ## warn if any entries in ... are not arguments for gam/gam.fit or gamm4/lmer 
     dots <- list(...)
     if(length(dots)){
-        validDots <- if(!is.na(fitter) && fitter=="gamm4"){
+        validDots <- if(!is.na(algorithm) && algorithm=="gamm4"){
                  c(names(formals(gamm4)), names(formals(lmer)))         
         } else {
             c(names(formals(gam)), names(formals(gam.fit)))
@@ -125,22 +168,22 @@ pffr <- function(
 	#start new formula
 	newfrml <- paste(responsename, "~", sep="")
 	newfrmlenv <- new.env()
-	evalenv <- if("data" %in% names(call)) eval(call$data) else NULL
+	evalenv <- if("data" %in% names(call)) eval.parent(call$data) else NULL
 	
 	nobs <- nrow(eval(responsename,  envir=evalenv, enclos=frmlenv))
 	nyindex <- ncol(eval(responsename,  envir=evalenv, enclos=frmlenv))
 	
 	
-	if(missing(fitter)||is.na(fitter)){
-		fitter <- ifelse(nobs > 1e5, "bam", "gam")
+	if(missing(algorithm)||is.na(algorithm)){
+		algorithm <- ifelse(nobs > 1e5, "bam", "gam")
 	} 
-	fitter <- as.symbol(fitter)
-	if(as.character(fitter)=="bam" && !("chunk.size" %in% names(call))){
+	algorithm <- as.symbol(algorithm)
+	if(as.character(algorithm)=="bam" && !("chunk.size" %in% names(call))){
 		call$chunk.size <- max(nobs/5, 10000) 
 		#same default as in bam
 	}
     ## no te-terms possible in gamm4
-	if(as.character(fitter)=="gamm4") stopifnot(length(where.te)<1)
+	if(as.character(algorithm)=="gamm4") stopifnot(length(where.te)<1)
     
 	
 	#if missing, define y-index or get it from first ff/sff-term, then assign expanded versions to newfrmlenv
@@ -179,15 +222,11 @@ pffr <- function(
 	assign(x=deparse(responsename), value=as.vector(t(eval(responsename, envir=evalenv, enclos=frmlenv))), 
 			envir=newfrmlenv)
     
-    #assign index values of missing y-values in _long_ format to .GlobaleEnv 
-    #s.t. smooth.construct.tensor.smooth.spec can access it later.
+   
     if(any(is.na(get(as.character(responsename), newfrmlenv)))){
-        assign(x=".PFFRmissingResponses", value=which(is.na(get(as.character(responsename), newfrmlenv))), 
-                envir=.GlobalEnv)     
-    }
-    ## FIXME: this is really ugly but since the impose.ffregC argument has to be propagated without evaluating
-    ## and smooth.construct.tensor.smooth.spec cannot access the model frame, I can't think of another way to do this.
-    ## The long and weird variable name should help avoid overwriting user-generated stuff in GlobalEnv. 
+        missingindname <- paste0(".PFFRmissingResponses.", paste(sample(letters, 6), collapse=""))
+        missingind <- which(is.na(get(as.character(responsename), newfrmlenv)))
+    } else missingind <- NULL
 	
 	##################################################################################
 	#modify formula terms.... 
@@ -289,13 +328,13 @@ pffr <- function(
         pcereterms <- lapply(pcreterms, function(x) x[names(x)!="data"])
     }else pcreterms <- NULL
     
-    #transform: s(x, ...), te(x, z,...), t2(x, z, ...) --> te(x, (z,)  yindex, ..., <bs.yindex>)
+    #transform: s(x, ...), te(x, z,...), t2(x, z, ...) --> <te|t2>(x, <z,> yindex, ..., <bs.yindex>)
     makeSTeT2 <- function(x){
         
         xnew <- x
-        if(deparse(x[[1]]) == "te" && as.character(fitter) == "gamm4") xnew[[1]] <- quote(t2)
+        if(deparse(x[[1]]) == "te" && as.character(algorithm) == "gamm4") xnew[[1]] <- quote(t2)
         if(deparse(x[[1]]) == "s"){
-            xnew[[1]] <- if(as.character(fitter) != "gamm4") {
+            xnew[[1]] <- if(as.character(algorithm) != "gamm4") {
                         tensortype 
                     } else quote(t2)
             #accomodate multivariate s()-terms
@@ -369,9 +408,19 @@ pffr <- function(
                                 names(lst[[i]])[length(names(lst[[i]]))] <- "impose.ffregC"
                                 lst[[i]][[llst+2]] <- nobs
                                 names(lst[[i]])[length(names(lst[[i]]))] <- "nobs"
+                                if(exists("missingindname")){
+                                    lst[[i]][[llst+3]] <- missingindname
+                                    names(lst[[i]])[length(names(lst[[i]]))] <- "missingindname"
+                                }
                             } else {
-                                lst[[i]] <- bquote(list(impose.ffregC = .(TRUE), 
-                                                nobs=.(nobs)))
+                                if(exists("missingindname")){
+                                    lst[[i]] <-   bquote(list(impose.ffregC = .(TRUE), 
+                                                    nobs=.(nobs),
+                                                    missingindname=.(missingindname)))
+                                } else {
+                                    lst[[i]] <- bquote(list(impose.ffregC = .(TRUE), 
+                                                    nobs=.(nobs)))
+                                }
                             }
                         }    
                         return(lst)
@@ -384,12 +433,26 @@ pffr <- function(
                                                         x$xt[[i]][[1]] == as.symbol("list") ||
                                                                 is.null(eval(x$xt[[i]][[1]])))))
                     xtra <- add.impose(x$xt)
-                    xtra[[length(xnew$d)+1]] <- bquote(list(impose.ffregC = .(TRUE), 
-                                    nobs=.(nobs)))
+                   
+                    append <- if(exists("missingindname")) {
+                         bquote(list(impose.ffregC = .(TRUE), 
+                                                nobs=.(nobs),
+                                                missingindname=.(missingindname)))
+                    } else {
+                        bquote(list(impose.ffregC = .(TRUE), 
+                                        nobs=.(nobs)))
+                    }
+                    xtra[[length(xnew$d)+1]] <- append
                     xtra
                 } else {
-                    imposearg <- bquote(list(impose.ffregC = .(TRUE), 
-                                            nobs=.(nobs)))
+                    imposearg <- if(exists("missingindname")) {
+                                bquote(list(impose.ffregC = .(TRUE), 
+                                                nobs=.(nobs),
+                                                missingindname=.(missingindname)))
+                            } else {
+                                bquote(list(impose.ffregC = .(TRUE), 
+                                                nobs=.(nobs)))
+                            }
                     
                     xtra <- bquote(list(.(imposearg)))
                     for (i in 3:(length(xnew$d)+1)) 
@@ -419,10 +482,11 @@ pffr <- function(
 	#... & assign expanded/additional variables to newfrmlenv
     where.notff <- c(where.c, where.par, where.s, where.te, where.t2)
 	if(length(where.notff)){
-        if("data" %in% names(call)) frmlenv <- list2env(eval(call$data), frmlenv) 
+        if("data" %in% names(call)) frmlenv <- list2env(eval.parent(call$data), frmlenv) 
         lapply(terms[where.notff], 
                 function(x){
                     #nms <- all.vars(x)
+              
                     if(any(unlist(lapply(terms[where.c], function(s) 
                                                 if(length(s)==1){
                                                     s==x
@@ -436,7 +500,7 @@ pffr <- function(
                     } 
                     ## remove names in xt, k, bs,  information (such as variable names for MRF penalties etc)
                     nms <- if(!is.null(names(x))){
-                                all.vars(x[names(x) == ""]) 
+                                all.vars(x[names(x) %in% c("", "by")]) 
                             }  else all.vars(x)
                     
                     
@@ -457,16 +521,16 @@ pffr <- function(
 	
 	newcall <- expand.call(pffr, call)
     newcall$yind <- newcall$tensortype <- newcall$bs.int <-
-            newcall$bs.yindex <- newcall$fitter <- NULL
+            newcall$bs.yindex <- newcall$algorithm <- NULL
 	newcall$formula <- newfrml
 	newcall$data <- quote(pffrdata)
-	newcall[[1]] <- fitter
+	newcall[[1]] <- algorithm
 	
     
     # add appropriate centering constraints for smooth effects
     # (not possible for gamm4, as the constraint destroys the simple 
     # diagonal structure of the t2-penalties)
-    if(!(as.character(fitter) %in% c("gamm4"))){
+    if(!(as.character(algorithm) %in% c("gamm4"))){
         suppressMessages(
                 trace(mgcv:::smooth.construct.tensor.smooth.spec, 
                         at = max(which(sapply(as.list(body(mgcv:::smooth.construct.tensor.smooth.spec)), function(x) any(grepl(x, pattern="object$C", fixed=TRUE))))) + 1, 
@@ -476,7 +540,7 @@ pffr <- function(
                                     
                                     if(!is.null(object$margin[[length(object$margin)]]$xt$impose.ffregC)){
                                         ## constraint: sum_i f(z_i, t) = 0 \forall t
-                                        cat("imposing constraints..\n")
+                                        #cat("imposing constraints..\n")
                                         nygrid <- length(unique(data[[object$margin[[length(object$margin)]]$term]]))
                                                                               
                                         
@@ -484,7 +548,8 @@ pffr <- function(
                                         Ctmp <- kronecker(t(rep(1, object$margin[[length(object$margin)]]$xt$nobs)), diag(nygrid))
                                         if(ncol(Ctmp) > nrow(object$X)){
                                             #drop rows for missing obs.
-                                            Ctmp <- Ctmp[, - get(".PFFRmissingResponses", .GlobalEnv)]
+                                            #Ctmp <- Ctmp[, - get(object$margin[[length(object$margin)]]$xt$missingindname, .GlobalEnv)]
+                                            Ctmp <- Ctmp[, -get("missingind", envir=sys.frame(3))]
                                         }
                                         C <- Ctmp %*% object$X
                                         
@@ -505,14 +570,15 @@ pffr <- function(
                                     if(!is.null(object$margin[[length(object$margin)]]$xt$impose.ffregC) &&
                                                             object$margin[[length(object$margin)]]$xt$impose.ffregC){
                                         ## constraint: sum_i f(z_i, t) = 0 \forall t
-                                        cat("imposing constraints..\n")
+                                        #cat("imposing constraints..\n")
                                         nygrid <- length(unique(data[[object$margin[[length(object$margin)]]$term]]))
                                        
                                         ## C = ((1,...,1) \otimes I_G) * B
                                         Ctmp <- kronecker(t(rep(1, object$margin[[length(object$margin)]]$xt$nobs)), diag(nygrid))
                                         if(ncol(Ctmp) > nrow(object$X)){
                                             #drop rows for missing obs.
-                                            Ctmp <- Ctmp[, - get(".PFFRmissingResponses", .GlobalEnv)]
+                                            #Ctmp <- Ctmp[, - get(object$margin[[length(object$margin)]]$xt$missingindname, .GlobalEnv)]
+                                            Ctmp <- Ctmp[, -get("missingind", envir=sys.frame(3))]
                                         }
                                         C <- Ctmp %*% object$X
                                         
@@ -534,9 +600,9 @@ pffr <- function(
     
     
     
-	# call fitter to estimate model
+	# call algorithm to estimate model
     m <- eval(newcall)
-    m.smooth <- if(as.character(fitter) %in% c("gamm4","gamm")){
+    m.smooth <- if(as.character(algorithm) %in% c("gamm4","gamm")){
         m$gam$smooth
     } else m$smooth
     
@@ -552,12 +618,18 @@ pffr <- function(
     lbls <- sapply(m.smooth, function(x) x$label)
     if(length(c(where.par, where.ffpc))){
         if(length(where.par)){
-            for(w in where.par) 
-                labelmap[[w]] <- {
-                    #covariates for parametric terms become by-variables:   
-                    where <- sapply(m.smooth, function(x) x$by) == names(labelmap)[w]
-                    sapply(m.smooth[where], function(x) x$label)
+            for(w in where.par){
+                # only combine if <by>-variable is a factor!
+                if(is.factor(get(names(labelmap)[w], envir=newfrmlenv))){
+                    labelmap[[w]] <- {
+                        #covariates for parametric terms become by-variables:   
+                        where <- sapply(m.smooth, function(x) x$by) == names(labelmap)[w]   
+                        sapply(m.smooth[where], function(x) x$label)
+                    }    
+                } else {
+                    labelmap[[w]] <- paste0("s(",yindvecname,"):",names(labelmap)[w])
                 }
+            }
         }
         if(length(where.ffpc)){
             ind <- 1
@@ -596,13 +668,15 @@ pffr <- function(
     }
     
     names(m.smooth) <- lbls
-    if(as.character(fitter) %in% c("gamm4","gamm")){
+    if(as.character(algorithm) %in% c("gamm4","gamm")){
         m$gam$smooth <- m.smooth 
     } else{
         m$smooth  <- m.smooth
     } 
     
-    ret <-  list(formula=formula, 
+    ret <-  list(
+            call=call,
+            formula=formula, 
             termmap=trmmap, 
             labelmap=labelmap, 
             responsename = responsename,
@@ -621,9 +695,10 @@ pffr <- function(
                     where.par=where.par
             ),
             ff=ffterms,
-            ffpc=ffpcterms)
+            ffpc=ffpcterms,
+            missingind = missingind)
     
-    if(as.character(fitter) %in% c("gamm4","gamm")){
+    if(as.character(algorithm) %in% c("gamm4","gamm")){
         m$gam$pffr <- ret
         class(m$gam) <- c("pffr", class(m$gam))
     } else {
